"""
Classes to handle run-metadata

"""
import os
import sys
import json
import copy
import platform
import traceback
import importlib
import csv
from dateutil import parser as dateparser
from datetime import datetime
from collections import defaultdict
import logging

logger = logging.getLogger('app')

class RunBase(object):
    """
    Base class for loading each run
    """
    def __init__(self, *args, **kwargs):
        """
        Initialize the Run class

        Args:

            config (object): Pipeline execution object
            rundir (str): Directory with runs
        """
        self.config = kwargs.pop('config')
        self.rundir = kwargs.pop('rundir')

        if not os.path.exists(self.rundir):
            raise Exception("missing rundir")

        self.files = []
        self.name = "Unknown"
        self.runid = "Unknown"
        self.load_time = datetime.now()
        self.deleted = False

    def __str__(self):
        return "{}.{}".format(self.name, self.runid)


    def load_metadata(self):
        """
        Load run metadata for a given rundir

        This will not load if the metadata size is more than
        1MB as a way to do sanity check.

        """
        metadatapath = os.path.join(self.rundir, 'metadata.json')
        if not os.path.exists(metadatapath):
            raise Exception("Could not locate metadata.json")

        statinfo = os.stat(metadatapath)
        if statinfo.st_size > 1024*1024: # 1MB
            raise Exception("Metadata file size too large")

        metadata = json.loads(open(metadatapath).read())
        self.metadata = metadata

        schema = metadata.get('schema', 'standard:pipeline')
        if schema not in ['standard:pipeline', 'standard:task']:
            raise Exception("Unknown run metadata structure")

        # Initialize some values. Have to switch to getattr
        self.start_time = dateparser.parse(metadata['start_time'])
        self.end_time = dateparser.parse(metadata['end_time'])

        duration = self.end_time - self.start_time
        self.duration = duration.days* 86400 + duration.seconds + 1000 * duration.microseconds

        self.status = metadata['status']
        self.name = metadata['name']
        config = metadata.get('config', {})
        self.display = metadata.get('display',
                                    config.get('display',self.name))
        self.description = metadata['description']
        self.computation = metadata.get('computation', 'batch')
        self.runid = metadata['runid']
        self.scheduled = metadata['scheduled']
        self.uid = metadata['uid']
        self.stats = metadata['stats']
        self.pid = metadata.get('pid', None)
        self.cmdline = metadata.get('cmdline', [])
        self.lineage = []

        # Patch for backward compatability
        if (('platform' in self.stats) and
            ('name' in self.stats['platform'])):
            self.stats['platform']['node'] = self.stats['platform']['name']


    def compute_lineage(self):
        """
        Compute lineage from the metadata
        """
        last  = {} # what is the last touch of a dataframe
        nodes = {}
        edges = []

        metadata = self.metadata
        if 'details' not in metadata:
            return

        # => First collect the entries
        allentries = []
        for framename in metadata['details']:

            details = metadata['details'][framename]
            description = details.get('description', '')

            ######################
            # => Sanity check...
            ######################
            if (('params' not in details) or
                (not isinstance(details['params'],list))):
                logger.error("Could not read params in metadata for {} in run {} metadata".format(framename, self.name))
                continue

            params = details['params']

            # Params is a list of inputs and outputs associated with the file.
            entries = [copy.copy(p) for p in params if (p.get('type', 'unknown') == 'lineage')]

            # Insert frame and description...
            for e in entries:
                e['frame'] = framename
                e['description'] = description

                # Insert timestamp if missing
                if 'timestamp' not in e:
                    if self.end_time is not None:
                        e['timestamp'] = self.end_time.timestamp()
                    else:
                        e['timestamp'] = self.start_time.timestamp()

            # Append to the global list.
            allentries.extend(entries)

        # => If there are any
        if len(allentries) == 0:
            return

        #=> Now process the entries...
        allentries = sorted(allentries, key=lambda entry: entry['timestamp'])

        for p in allentries:

            framename = p['frame']
            transform = p.get('transform','unknown')

            # => Anchor node
            core = "{} ({})".format(framename, transform)
            if core not in nodes:
                nodes[core] = {
                    'id': core,
                    'label': core.replace(" ","\n"),
                    'type': 'dataframe',
                    'text': "Dataframe\n" + p['description']
                }

            dependencies = p.get('dependencies', [])
            if len(dependencies) == 0:
                continue

            for d in dependencies:

                nature = d.get('nature', 'unknown').lower().strip()
                type_ = d.get('type', 'unknown').lower().strip()

                if type_ in ['column', 'columns']:
                    continue

                objects = d.get('objects', d.get('name', []))
                if not isinstance(objects, list) and not isinstance(objects, dict):
                    continue

                # Now clean and organize the dependencies...
                """
                    1. dependencies = [path1, path2..]
                    2. dependencies = {
                        "source1": path1,
                        "source2": description2
                    }
                    3. dependencies = {
                          "source1": {
                             "transform": "Trial",
                             "text": "...loan-s3/..."
                           }
                    }
                """
                cleaned_objects = {}
                if isinstance(objects, list):
                    # A list of files/text items
                    for o in objects:
                        label = os.path.basename(o) if "/" in o else o
                        cleaned_objects[label] = {
                            'transform': transform,
                            'text': str(o)
                        }
                elif isinstance(objects, dict):
                    for o, detail in objects.items():
                        # A dictionary of name: <object>
                        if isinstance(detail, str):
                            # A dictionary of name: str
                            cleaned_objects[o] = {
                                'transform': transform,
                                'text': detail
                            }
                        elif isinstance(detail, str):
                            # A dictionary of name: list
                            cleaned_objects[o] = {
                                'transform': transform,
                                'text': ",".join([str(s) for s in detail])
                            }
                        elif isinstance(detail, dict):
                            # A dictionary of name: dict
                            node_transform = detail.get('transform', transform)
                            if 'text' in detail:
                                text = detail['text']
                            else:
                                detail.pop('transform', None)
                                text = str(detail)
                            cleaned_objects[o] = {
                                'transform': node_transform,
                                'text': text
                            }

                for label, detail in cleaned_objects.items():
                    generator = detail['transform']
                    text = detail['text']
                    node = "{} ({})".format(label, generator)

                    if type_ in ['file']:
                        text = "Path/other: {}".format(text)
                    elif type_ in ['dataframe', 'dataframes']:
                        type_ = "dataframe"
                        node_transform = last.get(label, generator)
                        node = "{} ({})".format(label, node_transform)
                        text = "Dataframe: {}".format(text)
                    elif type_ in ['database', 'db', 'databases']:
                        type_ = "database"
                        node_transform = last.get(label, generator)
                        node = "{} ({})".format(label, node_transform)
                        text = "DB: {}".format(text)
                    elif type_ in ['api', 'url']:
                        type_ = "url"
                        node_transform = last.get(label, generator)
                        node = "{} ({})".format(label, generator)
                        text = "URL: {}".format(text)
                    elif type_ in ['columns', 'column']:
                        continue
                    else:
                        logging.warning("Unknown Type: {}".format(type_))
                        continue

                    # => Target node..
                    if node not in nodes:
                        nodes[node] = {
                            'id': node,
                            'label': node.replace(" ","\n"),
                            'type': type_,
                            'text': text
                        }

                    # => Add edge
                    source = core if nature == 'output' else node
                    target = node if nature == 'output' else core
                    edge = {
                        'id': 'link{}'.format(len(edges)+1),
                        'source': source,
                        'target': target
                    }

                    edges.append(edge)

            # => Keep track of what needs to be done.
            last[framename] = transform

        self.lineage = list(nodes.values()) + edges

    def get_lineage(self):
        """
        Return lineage in a way that it can be rendered by
        cytoscape.
        """
        return [ {'data': l} for l in self.lineage]

    def load(self):
        """
        Load the metadata and other aspects of run
        """
        self.load_metadata()
        try:
            self.compute_lineage()
            #lineage = self.get_lineage()
            #if len(lineage) > 0:
            #    print(json.dumps(lineage, indent=4))
        except:
            logger.exception("Unable to compute lineage for {}".format(str(self)))


    def compute_compliance_metadata(self):
        """
        Collect the column information from each of the frames
        """
        frames = {}

        metadata = self.metadata
        if 'details' not in metadata:
            return frames

        # => First collect the entries
        for framename in metadata['details']:
            """
            For each frame collect lineage and column information
            """

            details = metadata['details'][framename]
            description = details.get('description', "")
            frametype = details.get('frametype', "unknown")

            if (('params' not in details) or
                (not isinstance(details['params'],list))):
                continue

            params = details['params']

            # First check the finalized columns...
            columns_output = None
            columns_default = None
            inputs = {}
            outputs = {}

            for p in params:
                if not isinstance(p, dict):
                    continue
                _type = p.get('type', None)
                _action = p.get('action', None)
                _columns = p.get('columns', None)

                # => Have columns been specified, then use it.
                if _type == 'compute':
                    # Extract column information
                    if (isinstance(_columns,dict)) and (len(_columns) > 0):
                        if (_action == 'output'):
                            columns_output = _columns
                        else:
                            columns_default = _columns

                # Now collect information from lineage metadata
                if _type == 'lineage':

                    # Extract lineage information
                    dependencies = p.get('dependencies', [])
                    for d in dependencies:
                        nature = d.get('nature', 'unknown').lower().strip()
                        type_ = d.get('type', 'unknown').lower().strip()

                        if type_ in ['dataframe', 'columns']:
                            continue

                        # Dependencies can be specified as a
                        # (a) a list
                        # (b) a dict: name => str
                        # (c) a dict: name => list
                        # (d) a dict: name => dict
                        objects = d.get('objects',{})
                        if isinstance(objects, list):
                            for name in objects:
                                if not isinstance(name, str):
                                    continue
                                if nature == 'input':
                                    if name not in inputs:
                                        inputs[name] = name
                                    else:
                                        inputs[name] += " " + name
                                else:
                                    if name not in outputs:
                                        outputs[name] = name
                                    else:
                                        outputs[name] += " " + name
                        elif isinstance(objects, dict):
                            for name, desc in objects.items():
                                if isinstance(desc, list):
                                    desc = "; ".join(desc)
                                if nature == 'input':
                                    if name not in inputs:
                                        inputs[name] = desc
                                    else:
                                        inputs[name] += " " + desc
                                else:
                                    if name not in outputs:
                                        outputs[name] = desc
                                    else:
                                        outputs[name] += " " + desc

            # If column information has been collapsed, then use
            # it. Otherwise use the default column information
            if columns_output is not None:
                columns = columns_output
            else:
                columns = columns_default


            # => Now collect all the information in a single dict We
            # know for each frame, what the columns are, what was the
            # input and what was the output.
            frames[framename] = {
                'framename': framename,
                'frametype': frametype,
                'description': description,
                'type': frametype,
                'columns': columns,
                'inputs': inputs,
                'outputs': outputs
            }

        return frames

    def generate_compliance_summary(self):
        """
        Look through the dataframe metadata and
        lineage to determine
        """

        # Now we have to gather the columns
        frames = self.compute_compliance_metadata()

        summary = defaultdict(list)
        for frame in frames.values():

            # Consider only the outputs. Ignore the inputs
            # Each flow entry is <name-of-output> -> Text
            flows = list(frame['outputs'].items())

            #=> Collapse the information. There could potentially be multiple entries. How?
            io = {}
            for k,v in flows:
                if k not in io:
                    io[k] = []
                if isinstance(v, str):
                    io[k].append(v)
                elif isinstance(v, list):
                    io[k].extend(v)
                else:
                    io[k].append(str(v))

            for inputname, desc in io.items():
                if isinstance(desc, list):
                    desc = ". ".join(desc)
                if (('columns' not in frame) or
                    (frame['columns'] is None)):
                    continue
                for colname, coldetails in frame['columns'].items():
                    coldetails = copy.copy(coldetails)
                    coldetails['source_name'] = inputname
                    coldetails['source_description'] = desc + " " + frame['description']
                    coldetails['source_format'] = frame['frametype']
                    coldetails['column'] = colname
                    label = (inputname, colname)
                    summary[label].append(coldetails)


        # flatten list of coldetails...
        values = []
        for sublist in summary.values():
            for item in sublist:
                values.append(item)

        return values


    def load_files(self, include_output_frame=True):
        """
        Minimal loading of files
        """

        metadata = self.metadata
        if 'details' not in metadata:
            # Nothing to be done
            return

        for framename in metadata['details']:

            details = metadata['details'][framename]

            ######################
            # => Sanity check...
            ######################
            if (('params' not in details) or
                (not isinstance(details['params'],list))):
                logger.error("Could not read params in metadata for {} in run {} metadata".format(framename, self.name))
                continue

            if 'history' not in details:
                logger.error("Could not read history in metadata for {} in run {} metadata".format(framename, self.name))
                continue

            params = details['params']
            history = details['history']

            # Params is a list of inputs and outputs associated with the file.
            for p in params:

                if p.get('action', 'unknown') != 'output':
                    continue

                # Now we have params corresponding to a file write
                # action. Do we have a valid filename?
                if (('filename' not in p) or
                    (not isinstance(p['filename'], str)) or
                    (len(p['filename']) == 0)):
                    continue

                # A given frame can have multiple components, e.g.,
                # multiple partitions.
                if (('components' not in p) or
                    (not isinstance(p['components'], list)) or
                    (len(p['components']) == 0)):
                    continue

                # Get the first component only for simplicity
                component = p['components'][0]

                # Component looks like this...
                {
                    'filename': 'hopscotch/PLP/output/PLPTargetOnly/plptarget-20191002-165551/product_persona.csv',
                    'modified_time': 'Wed Oct 2 16:58:25 2019',
                    'sha256sum': 'e9731bdc22f545da71304e77d50da31ccb263458317a2c1b7b73177092ad5d73',
                    'columns': 2,
                    'filesize': '0.000 MB',
                    'create_time': 'Wed Oct 2 16:58:25 2019',
                    'rows': 1
                }

                root = self.config.enrich_data_dir
                path = os.path.join(root, component['filename'])
                frametype = p.get('frametype', 'pandas')

                detail = copy.copy(component)
                detail.update({
                    'name': framename,
                    'label': framename,
                    'path': path,
                    'frametype': frametype,
                })

                self.files.append(detail)

    def get_files(self):
        """
        Get the files loaded for this run

        Returns:
               files (list): List of dictionaries
        """
        return self.files

    def get_cmdline(self):
        """
        Get the cleaned command line used to generate the run
        """
        enrich_root = os.environ['ENRICH_ROOT']
        cmdline = self.cmdline
        cleaned = []
        for c in cmdline:
            if '.virtualenvs' in c:
                cleaned.append("..." +c[c.index('.virtualenvs') +
                                 len('.virtualenvs') + 1:])
            elif c.startswith(enrich_root):
                cleaned.append("..." +c[len(enrich_root) + 1:])
            else:
                cleaned.append(c)

        return cleaned

    def get_extra_args(self):
        """
        Get the extra args specified for this run
        """
        enrich_root = os.environ['ENRICH_ROOT']
        cmdline = self.cmdline
        extra = {}
        for c in cmdline:
            if ":" not in c or "=" not in c:
                continue

            module = c[:c.index(":")]
            name   = c[c.index(":")+1:c.index("=")]
            value  = c[c.index("=")+1:]

            if module not in extra:
                extra[module] = []
            extra[module].append({
                'name': name,
                'value': value
            })
        return extra

    def get_run_note(self):
        """
        Get any note given on the command line..
        """
        extra_args = self.get_extra_args()
        for arg in extra_args.get('global',[]):
            if arg['name'] == 'note':
                return arg['value']

        return ""

class RunManagerBase(object):
    """
    Load and keep all the runs ready
    """
    def __init__(self, *args, **kwargs):
        """
        Initialize the Run Manager

        Args:

             config (object): Pipeline execution object
             numruns (int): Number of runs to load

        """

        self.config = kwargs.pop('config')
        self.runs = []
        self.runcls = RunBase
        self.numruns = kwargs.get('numruns', 100)

    def load_one_run(self, outputdirs, runid):

        # Check if already present. If so, nothing to be done
        run = self.get_run(runid)
        if run is not None:
            run.load()
            return

        logger.debug("RunID ({})missing in cache. Loading".format(runid),
                     extra={
                         'transform': self.config.name
                     })
        rundir = None
        for outputdir in outputdirs:
            if runid in os.listdir(outputdir):
                rundir = os.path.join(outputdir, runid)
                break

        # Nothing to be done...
        if rundir is None:
            logger.error("RunID ({}) could not be found".format(runid),
                         extra={
                             'transform': self.config.name,
                             'data': 'Output dir(s): {}'.format(", ".join(outputdirs))
                         })
            return

        try:
            run = self.runcls(config=self.config,
                              name=runid,
                              rundir=rundir)
            run.load()
            self.runs.append(run)
            logger.debug("Loaded a new run: {}".format(runid),
                         extra={
                             'transform': self.config.name,
                         })
        except:
            logger.exception("Unable to load runid: {}".format(runid),
                             extra={
                                 'transform': self.config.name,
                             })
        return

    def load(self, offset=0, page=50, runid=None):
        """
        Load the runs

        Args:

             offset: offset from the last run
             page: number of runs per page
             runid: arbitrary runid (overrides the offset and page

        """

        # Default
        outputdirs = [self.config.outputdir]

        # Check if there is an archive (e.g., s3 mounted on a
        # path). If if exists then include it.
        backupconfig = self.config.siteconf.get('backup',{})
        mountroot = backupconfig.get('mount', None)
        if mountroot is not None:
            relpath = os.path.relpath(self.config.outputdir, self.config.enrich_root)
            mountdir = os.path.join(mountroot, relpath)
            if os.path.exists(mountdir):
                outputdirs.append(mountdir)

        if runid is not None:
            return self.load_one_run(outputdirs, runid)

        extra = 10
        run_cache = { r.runid : r for r in self.runs }
        self.runs = []

        runids = []
        count = 0
        for outputdir in outputdirs:

            #=> Check if the output dir exists
            if not os.path.exists(outputdir):
                continue

            # Default path where no run id is specified. If we have
            # not seen the outputdir, then include it.
            for runid in sorted(os.listdir(outputdir), reverse=True):
                if runid not in runids:
                    runids.append({
                        'runid': runid,
                        'rundir': os.path.join(outputdir, runid)
                    })
                    count += 1

                # Dont read much more...
                if count > offset + page + extra:
                    break

        errors = []
        for r in runids:
            if self.config.runid == r['runid']:
                # Dont read the current run's information. It wont be
                # there.
                continue

            runid = r['runid']
            rundir = r['rundir']

            # Get timestamp of the file.
            logfile = os.path.join(rundir, 'log.json')
            if not os.path.exists(logfile):
                continue

            try:
                mtime = os.path.getmtime(logfile)
            except:
                mtime = 0

            last_modified_date = datetime.fromtimestamp(mtime)

            # See if we have the latest copy. If so use it.
            if ((runid in run_cache) and
                (run_cache[runid].load_time > last_modified_date)):
                self.runs.append(run_cache[runid])
                continue

            # If it doesnt exist or the cope is not recent, the load it.
            try:
                run = self.runcls(config=self.config,
                                  name=runid,
                                  rundir=rundir)
                run.load()
                self.runs.append(run)
            except Exception as e:
                errors.append(str(r) + ": " + str(e))
                pass

        # Now that all the runs are loaded, sort them..
        self.runs = sorted(self.runs,
                           key=lambda r: r.start_time,
                           reverse=True)

        logger.debug("Loaded {} past runs".format(len(self.runs)),
                     extra=self.config.get_extra({
                         'transform': self.config.name,
                         'data': str(list([r.runid for r in self.runs]))
                     }))
        if len(errors) > 0:
            logger.error("Error while load {} past runs".format(len(errors)),
                         extra=self.config.get_extra({
                             'transform': self.config.name,
                             'data': errors
                         }))


    def get_run(self, runid):
        """
        Get run object for a given runid

        Args:

             runid (str): Run id

        """

        for r in self.runs:
            if r.runid == runid:
                return r

        return None

    def get_last_run(self, status="success"):
        """
        Get the last run with a given status

        if status is None or '' then return the
        last run irrespective of the status.

        Args:

             status (str): Status of the tune

        """
        for r in self.runs:
            if status in [None, 'any', '*', '']:
                return r

            if r.status == status and not r.deleted:
                return r
        return None

    def refresh(self):
        """
        Reload all the runs
        """
        self.runs = {}
        self.load()

class PipelineConfig(object):
    """
    Pipeline class. Simulates the pipeline execution
    service's internal object.
    """
    def __init__(self, conf):
        self.spec = conf
        self.name = conf['name']
        self.runid = "unknown"
        self.siteconf = {}
        self.set_params()

    def __str__(self):
        return self.name

    def set_params(self):

        # Set parameters
        params = {
            'name': self.spec['name'],
            'enrich_root': os.environ['ENRICH_ROOT'],
            'enrich_run_root': os.environ.get('ENRICH_RUN_ROOT',
                                              os.environ['ENRICH_ROOT']),
            'enrich_customers_dir': os.environ['ENRICH_CUSTOMERS'],
            'enrich_data_dir': os.environ['ENRICH_DATA']
        }

        for attr in ['data_root', 'output']:
            params[attr] = self.spec[attr] % params

        # => handle usecase_root/customer_root carefully for backward compatability
        usecase_root = self.spec.get('usecase_root',self.spec.get('customer_root', "Unknown"))
        params['usecase_root'] = params['customer_root'] = usecase_root % params

        # Handle aliases
        aliases = {
            'outputdir': 'output',
            'enrich_customers':  'enrich_customers_dir',
            'enrich_data': 'enrich_data_dir'
        }
        for k,v in aliases.items():
            params[k] = params[v]

        # Store them both as a
        self.params = params
        for k,v in self.params.items():
            setattr(self, k, v)

    def get_extra(self, extra):
        # Insert any additional attributes. Nothing as of now.
        return extra

def get_pipeline(path):
    """
    Take the given path and turn that into config. This
    function instantiates a PipelineConfig object and
    assigns it a run manager

    """
    try:
        if path.lower().endswith(".json"):
            confobj = json.load(open(path))
            confobj['conf'] = path
        else:
            # Load python
            modname = "".join([x if x.isalnum() else "_" for x in path])
            modspec = importlib.util.spec_from_file_location(modname, path)
            mod = importlib.util.module_from_spec(modspec)
            modspec.loader.exec_module(mod)
            if not hasattr(mod, 'config'):
                raise Exception("Invalid pipeline configuration path")
            confobj = mod.config
            confobj['conf'] = path
    except:
        logger.exception("Unable to load specified pipeline configuration")
        raise


    # => Now we have a dictionary of the configuration.
    config = PipelineConfig(confobj)

    # Add run manager
    config.runmanager = RunManagerBase(config=config)

    # Return
    return config



import os
import sys
import json
import copy
import re
import traceback
import logging

from ..lib.exceptions import *
from .node import *
from .mixins import *

logger = logging.getLogger('app')

__all__ = ['Source', 'Sink', 'Compute',
           'Trigger', 'NotificationIntegration',
           'DatabaseIntegration']

##################################################
# Transform objects
##################################################
class Source(Transform):
    """
    This is a special transform that introduces dataframes into the
    state.

    This transform typically extracts information from thirdparty
    tools and services such as APIs, databases.
    """
    def __init__(self, *args, **kwargs):
        super(Source,self).__init__(*args, **kwargs)
        self.roles_supported = ['Source']
        self.roles_current = 'Source'

    def sample_inputs(self):
        return []

    def is_source(self):
        return True

class Sink(Transform):
    """
    This is a special transform that dumps dataframes that are in the
    state to some thirdparty service such as file system, database,
    cloud storage etc.
    """
    def __init__(self, *args, **kwargs):
        super(Sink, self).__init__(*args, **kwargs)
        self.roles_supported = ['Sink']
        self.roles_current = 'Sink'

    def is_sink(self):
        return True

class Compute(Transform):
    """

    This is a typical transform that wrangles data. It could introduce
    new frames and transform existing frames.

    This will mainly be pandas code but may also be spark/other
    code. The framework itself is agnostic to what the transform
    does.

    """
    def __init__(self, *args, **kwargs):
        super(Compute,self).__init__(*args, **kwargs)
        self.roles_supported = ['Compute']
        self.roles_current = 'Compute'


    def add_marker(self, state, name=None, suffix="Completed"):
        """
        Adds an object to the state to force order. Doesnt
        serve any other purpose

        Parameters
        ----------
        state (obj): State object
        """
        detail = {
            'df': {},
            'transform': self.name,
            'frametype': 'dict',
            'params': {},
            'history': [],
        }

        # What should I call the state object?
        if name is None:
            name = self.name
            if suffix is not None:
                name += "_" + suffix

        # Dump it into the shared state
        state.update_frame(name, detail, create=True)


class Trigger(Sink):
    """
    This transform is expected to run after the computation
    is over. This is a particular kind of sync

    .. deprecated:: 2.0

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class NotificationIntegration(Integration):
    """
    Search interface dashboard

    .. deprecated:: 2.0

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tags.extend(['integration', 'notification'])

class DatabaseIntegration(Integration):
    """
    Write to database

    .. deprecated:: 2.0

    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tags.extend(['integration', 'database'])



import os
import json 
import logging
from  enrichsdk.core.node import Node

logger = logging.getLogger('app')

class Schema(object):
    """
    This class allows one to use load exported catalog data 
    """
    def __init__(self, filename):

        if not os.path.exists(filename):
            raise Exception("Invalid or missing schema file: {}".format(filename))

        try:
            catalogs = json.load(open(filename))
        except:
            logger.exception("Unable to read filename")
            raise Exception("Unsupported format: {}".format(filename))

        if isinstance(catalogs, dict):
            catalogs = [catalogs]
            
        for catalog in catalogs: 
            if (('schema' not in catalog) or
                (catalog['schema'] != 'Catalog')): 
                raise Exception("Only catalog level exports supported") 

        self.catalogs = catalogs 
            

    def find_source(self, catalogname, sourcename):

        catalog = None 
        for c in self.catalogs:
            if c['name'] == catalogname:
                catalog = c
                break 

        if catalog is None:
            return None

        source = None
        
        for s in catalog['sources']:
            if s['name'] == sourcename:
                source = s
                break 

        return source
    
    def get_column_docs(self, catalogname, sourcename):
        """
        Summarize the output in form that it usable by the 
        transform 
        """

        outputs = {}
        
        source = self.find_source(catalogname, sourcename)
        if source is None:
            # Didnt find the source 
            return outputs 
        
        for c in source['columns']:
            colname = c['name']
            coldesc = c.get('description', c.get('desc', ''))
            notes   = c.get('notes','')
            
            outputs[colname] = coldesc + "\n" + notes 

        return outputs 

class TransformSchemaMixin(object):

    def documentation_from_catalog(self, catalog, frames):
        
        # Check if the transform is an instance of the sdk
        if not isinstance(self, Node):
            raise Exception("Transform should be an instance of the Node")

        # Check if the catalog exists 
        catalog = self.config.get_file(catalog)
        schema = Schema(catalog)
        outputs = self.outputs
        
        for frame, details in frames.items():
            catalogname = details['catalog']
            sourcename = details['source'] 
            
            # Initialize 
            if frame not in outputs:
                outputs[frame] = {}
                
            frameoutputs = schema.get_column_docs(catalogname, sourcename)
            if len(frameoutputs) == 0:
                continue
            
            outputs[frame].update(frameoutputs) 
        
        



import os
import sys
import json
import re
import copy
from enrichsdk import Sink
import logging

logger = logging.getLogger("app")

class JSONSink(Sink):
    """
    Store a 'dict' frame that is present in the state into a file.

    Params are meant to be passed as parameter to update_frame.

    Example configuration::

         "args": {
             "sink": {
                 'test': {
                     'frametype': 'dict',
                     'filename': '%(output)s/%(runid)s/mytestoutput.json',
                     'params': {}
                 }
             }
         }
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = "JSONSink"

        self.testdata = {
            'conf': {
                'args': {
                    'sink': {
                        'frame1': {
                            'frametype': 'dict',
                            'filename': '%(output)s/%(runid)s/mytestoutput.json',
                            'params': {}
                        }
                    }
                }
            },
            'data': {
                'frame1': {
                    'filename': 'outputjson.json',
                    'frametype': 'dict',
                    'transform': 'TestJSON',
                    'params': {
                    }
                }
            }
        }


    def preload_clean_args(self, args):
        """
        Clean when the spec is loaded...
        """

        if 'sink' not in args:
            args = {
                'sink': args
            }

        args = super().preload_clean_args(args)

        assert 'sink' in args
        assert isinstance(args['sink'], dict)
        assert len(args['sink']) > 0

        sink = args['sink']
        for name, detail  in sink.items():

            if (("frametype" not in detail) or
                (detail['frametype'] != 'dict')):
                logger.error("Invalid configuration. Only JSON/Dictionaries are supported by this sink transform",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            if (('filename' not in detail) or
                (not isinstance(detail['filename'], str)) or
                ('params' not in detail) or
                (not isinstance(detail['params'], dict))):
                logger.error("Invalid args. Filename (string) and params (dict) are required",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            filename = detail['filename']
            if not filename.lower().endswith('.json'):
                logger.error("Input file must a .json file",
                             extra=self.config.get_extra({
                                 'transform': self.name
                             }))
                raise Exception("Invalid configuration")

            detail['root'] = self.config.enrich_data_dir

            tags = detail.get('tags', [])
            if isinstance(tags, str):
                tags = [tags]
            detail['tags'] = tags

            #=> Materialize the path...
            detail['filename'] = self.config.get_file(detail['filename'],
                                                      extra={
                                                          'frame_name': name
                                                      })

        return args

    def validate_args(self, what, state):
        """
        An extra check on the arguments to make sure
        it is consistent with the specification
        """
        args = self.args
        assert 'sink' in args
        assert isinstance(args['sink'], dict)
        assert len(args['sink']) > 0

        sink = args['sink']
        for name, detail  in sink.items():
            assert (('frametype' in detail) and (detail['frametype'] == 'dict'))
            assert 'filename' in detail
            assert 'params' in detail

    def process(self, state):
        """
        Store the dictionary 'frames' in the state in files.
        """
        logger.debug("{} - process".format(self.name),
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))


        available_frames = state.get_frame_list()

        # => First construct input for the pandasframe
        extra = {}
        args_input = {}
        write_input = {}
        framecls = self.config.get_dataframe('dict')

        sink = self.args['sink']
        for pattern in sink:
            # The pattern could be precise dataframe name or could be
            # regular expression.
            regex = re.compile('^{}$'.format(pattern))
            frames = [m.group(0) for f in available_frames for m in [regex.search(f)] if m]
            if len(frames) == 0:
                logger.warning("Pattern has not matched any frames: {}".format(pattern))
                continue

            for f in frames:

                # Get the details of this frame
                detail = state.get_frame(f)

                # Handle frametype
                frametype = detail['frametype']
                if frametype != 'dict':
                    logger.warning("Pattern has matched non-dict frame: {}".format(f),
                                   extra=self.config.get_extra({
                                       'transform': self.name
                                   }))
                    continue

                # Now construct the output file name
                filename = sink[pattern]['filename']
                filename = self.config.get_file(filename,
                                                create_dir=True,
                                                extra={
                                                    'frame_name': f
                                                })

                extra[f] = {
                    'notes': self.collapse_notes(detail),
                    'descriptions': self.collapse_descriptions(detail)
                }

                params = sink[pattern].get('params',{})
                write_input[f] = {
                    'frametype': detail['frametype'],
                    'filename': filename,
                    'pattern': pattern,
                    'df': detail['df'],
                    'params': params
                }

                args_input[f] = copy.copy(sink[pattern])
                args_input[f]['filename'] = filename

        framecls.write(args_input, write_input)

        for name in write_input:

            detail = write_input[name]

            # => Insert columns and tags
            pattern = detail['pattern']
            detail['params']['tags'] = sink[pattern]['tags']

            # Incorporate columns, notes and description
            detail['params'].update(extra[name])

            detail['params'] = [
                detail['params'],
                {
                    'type': 'lineage',
                    'transform': self.name,
                    'dependencies': [
                        {
                            'type': 'dataframe',
                            'nature': 'input',
                            'objects': [
                                name
                            ]
                        },
                        {
                            'type': 'file',
                            'nature': 'output',
                            'objects': [
                                detail['params']['filename']
                            ]
                        }
                    ]
                }
            ]

            # Insert additional detail
            detail['transform'] = self.name
            detail['history'] = [
                {
                    'transform': self.name,
                    'log': "Wrote output"
                }
            ]

            state.update_frame(name, detail)

        logger.debug("Finished writing data",
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """
        pass


provider = JSONSink



import logging
from logging.config import dictConfig

logconfig = {
    'version': 1,
    "disable_existing_loggers": False,
    'formatters': {
        "basic": {
            "class": "logging.Formatter",
            "datefmt": "%I:%M:%S",
            "format": "%(asctime)s %(levelname)s %(name)s %(message)s"
        },
        'expanded': {
            "class": "logging.Formatter",
            "format": "%(asctime)s - %(name)s - %(transform)s - %(levelname)s - %(message)s%(data)s",
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'basic',
            'level': 'INFO',
            "stream": "ext://sys.stdout"
        },
        'app': {
            'class': 'logging.StreamHandler',
            'formatter': 'expanded',
            'level': 'DEBUG',
            "stream": "ext://sys.stdout"
        }
    },
    'loggers': {
        '': {
            'level': 'INFO',
            'handlers': ['console'],
            'propagate': False
        },
        'default': {
            'level': 'INFO',
            'handlers': ['console'],
            'propagate': False
        },
        'app': {
            'level': 'DEBUG',
            'handlers': ['app'],
            'propagate': False
        },
    },
    'root': {
        'level': 'WARN',
        'handlers': ['console']
    },
}

def setup_logging():
    dictConfig(logconfig)

    for name in [
            'kazoo.client',
            'pykafka.simpleconsumer',
            'pykafka.handlers',
            'pykafka.balancedconsumer',
            'pykafka.connection',
            'pykafka.membershipprotocol',
            'pykafka.cluster',
            'pykafka.topic',
            'matplotlib.pyplot',
            'asyncio'
    ]:
        logging.getLogger(name).disabled = True



"""
Metadata Format for FeatureStore
"""
import os
import sys
import json
from datetime import datetime, timedelta
from collections import OrderedDict

class FeatureGroupBase:
    """
    Customizable metadata class
    """
    def __init__(self, *args, **kwargs):

        self.schema = kwargs.pop('schema', "unknown:enrich")
        self.attributes = []
        self.elements = {}

    def validate(self):
        """
        Check if the required attributes are present, have
        the right format, and content
        """

        errors = []

        # Attr may be present or absent
        for attr in self.attributes:
            if attr.get('required', True):
                if attr['name'] not in self.elements:
                    errors.append("%(name)s is missing" % attr)
                    continue
            if attr['name'] not in self.elements:
                continue

            value = self.elements[attr['name']]
            if 'type' in attr:
                if not isinstance(value, attr['type']):
                    errors.append("%(name)s has wrong datatype" % attr)
                    continue

            # Validate function can be explicitly or implicitly specified
            validatefunc = None
            if 'validate' in attr:
                validatefunc = attr['validate']
            elif hasattr(self, 'validate_' + attr['name']):
                validatefunc = getattr(self, 'validate_' + attr['name'])

            if validatefunc is not None:
                if not callable(validatefunc):
                    errors.append("%(name)s validate function is not callable" % attr)
                    continue

                try:
                    vresult, verrors = validatefunc(attr, value)
                    if not vresult:
                        if isinstance(verrors, str):
                            errors.append(verrors)
                        elif isinstance(verrors, list):
                            errors.extend(verrors)
                        else:
                            errors.append(str(verrors))
                        continue

                except Exception as e:
                    errors.append("{} validate function has error: {}".format(attr['name'],
                                                                                str(e)))
                    continue

        return len(errors) == 0, errors

    def export(self, dummy=False):
        """
        Export an ordered dictionary with the content
        """
        response = OrderedDict([])
        response['schema'] = self.schema
        for attr in self.attributes:
            required = attr.get('required', True)
            name = attr['name']

            if dummy:
                response[attr['name']] = attr['default']
                continue

            if name not in self.elements:
                if not required:
                    continue
                else:
                    raise Exception("Incomplete metdata. Missing: {}".format(name))

            value = self.elements[attr['name']]
            if hasattr(value, 'serialize'):
                value = value.serialize()

            try:
                json.dumps(value)
            except:
                raise Exception("Unable to serialize")

            response[attr['name']] = value

        return response

    def add(self, attrname, value):
        """
        Add an attribute
        """
        selected = None
        for attr in self.attributes:
            if attrname != attr['name']:
                selected = attr
                continue

        if selected is None:
            raise Exception("Unknown attribute name specified")

        attrtype = selected.get('type', str)
        if not isinstance(attrtype, list):
            self.elements[attrname] = value
        else:
            # A list
            if attrname not in self.elements:
                self.elements[attrname] = []

            if isinstance(value, list):
                self.elements[attrname].extend(value)
            else:
                self.elements[attrname].append(value)


    def validate_simple_dict(self, attr, value):

        if not isinstance(value, dict):
            return False, "Invalid metadata type"

        if len(value) == 0:
            return False, "Empty metadata"

        return True, ""


class FeatureGroupSpec(FeatureGroupBase):
    """
    Specification of the featuregroup

    Required fields:
       name: Name of the feature group,
       description: Description of the feature group
       owner: email address or other identity (str),
       version: version of this feature group (str)
       join_keys: list of strings
       execution_spec: Uninterpreted dictionary



    Optional field: run
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, schema="v1:spec:featurestore:enrich", **kwargs)
        self.attributes = []

        # Collect all the attributes
        defaults = {
            'name':  "online.customer_persona.buyer",
            'description': "Customer attributes",
            'owner': "john@acmeinc.com",
            'version': 'v1.2'
        }
        for attr, default in defaults.items():
            self.attributes.append({
                'name': attr,
                'default': default,
                'type': str
            })

        for attr in ['joinkeys']:
            self.attributes.append({
                'name': attr,
                'type': list,
                'default': ['customer_id', 'division_id'],
                'required': True,
            })

        for attr in ['executionspec']:
            self.attributes.append({
                'name': attr,
                'type': dict,
                'required': True,
                'default': {
                    "cron": "0 20/1 * * * 2",
                    "pipeline": {
                        "name": "SortIdentifier"
                    }
                },
                'validate': self.validate_simple_dict
            })

    def validate_joinkeys(self, attr, value):

        if not isinstance(value, list):
            return False, "Invalid joinkeys"

        if len(value) == 0:
            return False, "Empty join keys"

        for v in value:
            if not instance(v, str) or len(v) == 0:
                return False, "Invalid/empty joinkeys: {}".format(str(value))

        return True, ""

class FeatureGroupRun(FeatureGroupBase):
    """
    Data about featuregroup runs
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, schema="v1:run:featurestore:enrich", **kwargs)
        self.attributes = []

        defaults = {
            'featuregroup_id': 3,
            "runid": "customer-291712-291122",
            "path": "enrich-acme/customer_persona/buyer/2020-01-02/data.pq",
            "run_start": (datetime.now() + timedelta(seconds=-3600*2.5)).isoformat()
        }

        # Collect all the attributes
        for attr, default in defaults.items():
            self.attributes.append({
                'name': attr,
                'default': default,
                'type': str
            })

        for attr in ['run_end']:
            self.attributes.append({
                'name': attr,
                'type': str,
                "default": (datetime.now() + timedelta(seconds=-3600*1)).isoformat(),
                'required': False,
            })

        for attr in ['columns']:
            self.attributes.append({
                'name': attr,
                'type': list,
                'default': [
                    {
                        'name': 'buyer_id',
                    },
                    {
                        'name': 'cm',
                    },
                    {
                        'name': 'ltv',
                    },
                    {
                        'name': 'first_date',
                    },
                    {
                        'name': 'last_date'
                    }
                ],
                'required': True,
            })

        for attr in ['metadata']:
            self.attributes.append({
                'name': attr,
                'type': dict,
                'required': True,
                'default': {
                    "execution": {
                        "description": "Experimental pipeline for testing transforms",
                        "cmdline": [
                            "/home/pingali/.virtualenvs/dev/bin/enrichpkg",
                            "test-transform",
                            "--capture",
                            "transforms/PLPSortMain/"
                        ],
                        "host": "whale",
                        "pid": 3043,
                        "runid": "exp-pipeline-20180922-192321",
                        "name": "ExperimentalPipeline",
                        "end_time": "2019-09-16T15:11:32",
                        "start_time": "2019-09-16T15:11:32",
                        "stats": {
                            "platform": {
                                "distribution": [
                                    "Ubuntu",
                                    "16.04",
                                    "xenial"
                                ],
                                "node": "whale",
                                "processor": "x86_64",
                                "release": "4.15.0-60-generic",
                                "python": "3.5.2",
                                "os": "Linux"
                            },
                            "user": "pingali"
                        }
                    }
                },
                'validate': self.validate_simple_dict
            })

    def validate_columns(self, attr, value):

        valid, message = super().validate_simple_dict(attr, value)
        if not valid:
            return valid, message

        for v in value:
            if not instance(v, dict) or len(v) == 0:
                return False, "Invalid/empty columns: {}".format(attr['name'])

            if 'name' not in v:
                return False, "Invalid/empty columns: {}".format(attr['name'])

        return True, ""




"""
File Operations
^^^^^^^^^^^^^^^

"""
import os
import sys
import abc
import shutil
import numpy as np
import pandas as pd
from enrichsdk import Trigger
from datetime import datetime
import logging

logger = logging.getLogger("app")

class FileOperationsBase(Trigger):
    """
    Base class for a FileOperations transform. For now only
    one action is supported 'copy'. More actions will be
    added in future.

    Example::

	    {
		"transform": "FileOperations",
		"enable": true,
		"dependencies": {
                   ....
		},
		"args": {
		    "actions": [
			{
			    "action": "copy",
			    "src": "%(output)s/%(runid)s/profile.sqlite",
			    "dst": "%(data_root)s/shared/campaigns/profile_daily/profile.sqlite",
			    "backupsuffix": ".backup"	
			},
                     ]
                }
            }

    """

    @classmethod
    def instantiable(cls):
        return False

    def __init__(self, *args, **kwargs):
        super(FileOperationsBase,self).__init__(*args, **kwargs)
        self.name = "FileOperationsBase"
        self.outputs = {}
        self.dependencies = {}

    def preload_clean_args(self, args):
        """
        Clean when the spec is loaded...
        """

        # Update the args
        args = super().preload_clean_args(args)

        # Sanity check...
        if not isinstance(args, dict):
            raise Exception("args should be a dictionary")

        if (('actions' not in args) or
            (not isinstance(args['actions'], list))):
            raise Exception("actions is missing or invalid")

        for a in args['actions']:
            if not isinstance(a, dict):
                raise Exception("Each action spec should be a dictionary")
            supported = ['copy']
            if a['action'] not in supported:
                raise Exception("Unsupported action")

            if a['action'] == 'copy':
                if (('src' not in a) or ('dst' not in a)):
                    raise Exception("Each copy action spec should specify a src and dst")

        return args

    def process(self, state):
        """
        Run the computation and update the state
        """
        logger.debug("{} - process".format(self.name),
                     extra=self.config.get_extra({
                         'transform': self.name
                     }))

        msg = ""
        actions = self.args['actions']
        for a in actions:
            action = a['action']

            # Pass any global variables...
            srcbase = self.config.get_file(a['src'], extra=self.args)
            dstbase = self.config.get_file(a['dst'], extra=self.args)

            if 'files' not in a:
                copyactions = [
                    {
                        'src': srcbase,
                        'dst': dstbase
                    }
                ]
            else:
                copyactions = []
                for f in a['files']:
                    copyactions.append({
                        'src': os.path.join(srcbase, f),
                        'dst': os.path.join(dstbase, f)
                    })

            backupsuffix = a.get('backupsuffix', ".backup")
            data_root = self.config.get_file("%(enrich_data_dir)s")
            for ca in copyactions:
                src = ca['src']
                dst = ca['dst']

                if not os.path.exists(src):
                    raise Exception("Could not find source: {}".format(src))

                if os.path.exists(dst):
                    backupsuffix = datetime.now().strftime(backupsuffix)
                    backupdst = dst + backupsuffix
                    if os.path.exists(backupdst):
                        if os.path.isdir(backupdst):
                            shutil.rmtree(backupdst)
                        else:
                            os.remove(backupdst)
                    os.rename(dst, dst + backupsuffix)

                try:
                    os.makedirs(os.path.dirname(dst))
                except:
                    pass

                # Handle the directory names
                if os.path.isdir(src):
                    shutil.copytree(src, dst)
                else:
                    shutil.copy(src, dst)

                msg += "Copy: {} => {}\n".format(os.path.relpath(src, data_root),
                                                 os.path.relpath(dst, data_root))

        logger.debug("{} - Completed".format(self.name),
                     extra=self.config.get_extra({
                         'transform': self.name,
                         'data': msg
                     }))

        ###########################################
        # => Return
        ###########################################
        return state

    def validate_results(self, what, state):
        """
        Check to make sure that the execution completed correctly
        """
        pass


provider = FileOperationsBase



import copy
import json
import importlib
import pandas as pd

############################################
# Pandas add a simple helper function
############################################
def _get_generic_dtype(self, colname):
    d = self.dtypes[colname]
    d = str(d)
    if ((d not in ['int64', 'float64', 'int8', 'bool', 'category']) and
        (not d.startswith('datetime'))):
        d = 'str'
    return d
pd.DataFrame.get_generic_dtype = _get_generic_dtype

############################################
# Subclass to Add Metadata
############################################
class PandasDFWithMetadata(pd.DataFrame):
    """
    Custom subclass of PandasDataFrame to support
    metadata collection.
    """

    _metadata = ['enrich_attrs']

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.enrich_attrs = {
            'name': '',
            'lineage': []
        }

    @property
    def _constructor(self):
        return PandasDFWithMetadata

    @property
    def lineage(self):
        return self.enrich_attrs.get('lineage',[])

    def set_lineage(self, value):
        self.enrich_attrs['lineage'] = value

    def clear_lineage(self):
        self.enrich_attrs['lineage'] = []

    def derive(self, inputcols, outputcols, inputname=None):
        """
        Note how a new column is derived from an existing
        columns from the same or different dataframe with
        a name
        """

        if isinstance(inputcols, str):
            inputcols = [inputcols]
        if isinstance(outputcols, str):
            outputcols = [outputcols]

        outputname = self.enrich_attrs.get('name', None)

        self.enrich_attrs['lineage'].append({
            'outputname': outputname,
            'inputname': inputname,
            'input': inputcols,
            'output': outputcols
        })

    def derive_from_map(self, mapping):
        for old, new in mapping.items():
            self.derive([old], [new])

    def get_lineage(self):
        return self.enrich_attrs['lineage']


    # https://github.com/pandas-dev/pandas/issues/2485
    def _combine_frame(self, other, *args, **kwargs):
        return super()._combine_frame(other, *args, **kwargs).__finalize__(self)

    # Overrides the rename pandas function.
    def rename(self, *args, **kwargs):

        columns = None
        if (('columns' in kwargs) and
            (isinstance(kwargs['columns'], dict))):
            columns = copy.copy(kwargs['columns'])
        value = super().rename(*args, **kwargs)
        if columns is not None:
            self.derive_from_map(columns)

        return value

spec = importlib.util.find_spec("pyspark")
if spec is not None:
    """
    Handle the case wheen pyspark is not installed on the localmachine
    """
    from pyspark import SparkContext
    from pyspark.sql import Column
    from pyspark.sql.functions import col

    #
    # https://stackoverflow.com/questions/46667810/how-to-update-pyspark-dataframe-metadata-on-spark-2-1
    def withMeta(self, alias, meta):
        """
        Add metadata to spark RDD


         Example::

             # new metadata:
             meta = {"ml_attr": {"name": "label_with_meta",
                                 "type": "nominal",
                                 "vals": [str(x) for x in range(6)]}}

             df_with_meta = df.withColumn("label_with_meta", col("label").withMeta("", meta))

        """
        sc = SparkContext._active_spark_context
        jmeta = sc._gateway.jvm.org.apache.spark.sql.types.Metadata
        return Column(getattr(self._jc, "as")(alias, jmeta.fromJson(json.dumps(meta))))

    Column.withMeta = withMeta




"""
Exceptions generated by the various modules 

"""
class InvalidConfiguration(Exception):     
    """
    Configuration error. 
    
    Raise this for invalid conf,args and test data. 
    """
    def __init__(self, message, elements): 
        super(InvalidConfiguration, self).__init__(message) 
        self.elements = elements 

    def __str__(self): 
        s = super(InvalidConfiguration, self).__str__()
        return s + ": " + str(self.elements) 

class InvalidTransform(Exception):     
    """
    Invalid transform. 

    Raise this from Transform and Model modules. 
    """
    def __init__(self, message, elements=[]): 
        super(InvalidTransform, self).__init__(message) 
        self.elements = elements         

class InvalidPlugin(Exception):
    """
    Invalid plugin

    Raise this from generic non-transform plugins
    """    
    def __init__(self, message, elements=[]): 
        super(InvalidPlugin, self).__init__(message) 
        self.elements = elements
        
class InvalidSkin(Exception):     
    """
    Invalid Skin

    Raise this from skin module
    """
    def __init__(self, message, elements=[]): 
        super(InvalidSkin, self).__init__(message) 
        self.elements = elements         

class Unsupported(Exception):     
    """
    Unsupported functionality 

    Raise this if there is missing configuration or unsupported
    conbombination of parameters.
    """
    def __init__(self, message, elements=[]): 
        super(Unsupported, self).__init__(message) 
        self.elements = elements         

class InvalidNode(Exception):
    """
    Could not find a transform/model/other subclass of 
    Node with a given name
    """
    pass


        



import logging
import sys
from logstash_formatter import LogstashFormatterV1

def log_setup(logfile=None):


    if logfile is not None:
        handler = logging.handlers.RotatingFileHandler(logfile,
                                                       maxBytes=10*1000*1000,
                                                       encoding='utf-8')
    else:
        handler = logging.StreamHandler(stream=sys.stdout)

    handler.setFormatter(LogstashFormatterV1())
    logging.basicConfig(handlers=[handler], level=logging.DEBUG)

    # Disable some handlers
    for name in [
            'kazoo.client',
            'pykafka.simpleconsumer',
            'pykafka.handlers',
            'pykafka.balancedconsumer',
            'pykafka.connection',
            'pykafka.membershipprotocol',
            'pykafka.cluster',
            'pykafka.topic',
            'py4j.java_gateway'
    ]:
        logging.getLogger(name).disabled = True



import traceback 
from functools import update_wrapper
import click
import logging

logger = logging.getLogger('app')

def log_activity(runfunc):
    """
    This is a decorator that wraps around a click command. 
    
    This decorator captures run metadata, initiates logging,
    and handles exceptions during execution of the command. 

    Example::

        @click.command('compare')
        @log_activity
        @click.pass_obj
        def _compare(config):
          "Compares the data" 
           ....
    
    """
    @click.pass_context 
    def logged_run(ctx, *args, **kwargs):
        """
        Make note of what is going in and out...
        """

        config = ctx.obj
        config.start_run(ctx.command.name) 
        try:
            try:
                config.state['command']['name'] = ctx.command_path
                config.state['command']['description'] = ctx.command.help
            except:
                traceback.print_exc() 
                pass

            response = ctx.invoke(runfunc, *args, **kwargs)
            config.set_status('success')
            config.end_run()
            return response 
        except Exception as e: 
            config.set_message(str(e))
            config.set_status('failure')
            logger.exception("Failure in execution")
            config.end_run()
            print("Command failed. Please see the log") 

    return update_wrapper(logged_run, runfunc)
    
